package com.gnarly.game.objects;

import static org.lwjgl.glfw.GLFW.GLFW_KEY_A;
import static org.lwjgl.glfw.GLFW.GLFW_KEY_D;
import static org.lwjgl.glfw.GLFW.GLFW_KEY_LEFT_SHIFT;
import static org.lwjgl.glfw.GLFW.GLFW_KEY_S;
import static org.lwjgl.glfw.GLFW.GLFW_KEY_SPACE;

import com.gnarly.engine.components.Animation;
import com.gnarly.engine.components.Shader;
import com.gnarly.engine.components.VAO;
import com.gnarly.engine.display.Camera;
import com.gnarly.engine.display.Window;
import com.gnarly.engine.utils.Library;
import com.gnarly.engine.utils.RHitbox;

public class Character {

	private final boolean
		LEFT = true,
		RIGHT = false;
	
	private final int
		IDLE = 0,
		WALK = 1,
		RUN = 2,
		JUMP = 3,
		FALL = 4;
	
	private final float
		WALK_ACCEL = 1,
		SPRINT_ACCEL = 2,
		STRAFE_ACCEL = 0.05f,
		WALK_SPEED = 1,
		SPRINT_SPEED = 2,
		STRAFE_SPEED = 0.8f,
		GROUND_DECAY = 0.15f,
		AERIAL_DECAY = 0.01f,
		FASTFALL = 2.0f,
		BOUNCE_SPEED = 6.0f,
		BOUNCE_SCALE = 0.8f;
	
	private final float
		GRAVITY = 0.1f,
		MAX_GRAV = 8.0f;
	
	private final float
		SHORT_HOP = 0.65f,
		LONG_JUMP = 0.35f,
		AERIAL_JUMP = 3.0f;
	
	private final int
		NUM_JUMPS = 50,
		MIN_JUMP = 4,
		MAX_JUMP = 7;
	
	private float fallSpeed;
	
	private int state = 0;
	
	private float x, y, width, height, dx, dy;
	
	private boolean airborne = false;
	private boolean dir = false;

	private int jump = 0;
	private int jumps = 0;
	private int jumpFrames = 0;
	
	private Animation anims[];
	private Shader shader;
	private VAO right;
	private VAO left;
	private Camera camera;
	private Window window;
	private RHitbox rHitbox;
	
	public Character(Window window, Camera camera, String name, float x, float y) {
		this.camera = camera;
		this.window = window;
		anims = new Animation[5];
		anims[IDLE] = new Animation(name + " idle", "png", 8, 10, true);
		anims[WALK] = new Animation(name + " walk", "png", 9, 10, true);
		anims[RUN] = new Animation(name + " run", "png", 4, 10, true);
		anims[JUMP] = new Animation(name + " jump", "png", 6, 20, true);
		anims[FALL] = new Animation(name + " fall", "png", 6, 20, true);
		this.x = x;
		this.y = y - anims[0].getHeight();
		width = anims[0].getWidth();
		height = anims[0].getHeight();
		shader = Library.getShader("default");
		rHitbox = new RHitbox(this.x, this.y, width, height);
		float[] vertices = new float[] {
			0.0f,  0.0f,   0.5f, //TOP LEFT
			0.0f,  height, 0.5f, //BOTTOM LEFT
			width, height, 0.5f, //BOTTOM RIGHT
			width, 0.0f,   0.5f  //TOP RIGHT
		};
		int[] indices = new int[] {
			0, 1, 3,
			1, 2, 3
		};
		float[] texCoords = new float[] {
			0, 0,
			0, 1,
			1, 1,
			1, 0
		};
		right = new VAO(vertices, indices, texCoords);
		texCoords = new float[] {
			1, 0,
			1, 1,
			0, 1,
			0, 0
		};
		left = new VAO(vertices, indices, texCoords);
	}
	
	public void update() {
		boolean move = false;
		if(window.isKeyPressed(GLFW_KEY_D)) {
			move = true;
			if(window.isKeyPressed(GLFW_KEY_LEFT_SHIFT) && !airborne && dx < 2) {
				if(dx > 0)
					dx += SPRINT_ACCEL;
				else
					dx += SPRINT_ACCEL / 20;
				if(dx > SPRINT_SPEED)
					dx = SPRINT_SPEED;
			}
			else if(!airborne && dx < 1) {
				if(dx > 0)
					dx += WALK_ACCEL;
				else
					dx += WALK_ACCEL / 20;
				if(dx > WALK_SPEED)
					dx = WALK_SPEED;
			}
			else if(dx < STRAFE_SPEED)
				dx += STRAFE_ACCEL;
		}
		if(window.isKeyPressed(GLFW_KEY_A)) {
			move = true;
			if(window.isKeyPressed(GLFW_KEY_LEFT_SHIFT) && !airborne && dx > -2) {
				if(dx < 0)
					dx -= SPRINT_ACCEL;
				else
					dx -= SPRINT_ACCEL / 20;
				if(dx < -SPRINT_SPEED)
					dx = -SPRINT_SPEED;
			}
			else if(!airborne && dx > -1) {
				if(dx < 0)
					dx -= WALK_ACCEL;
				else
					dx -= WALK_ACCEL / 20;
				if(dx < -WALK_SPEED)
					dx = -WALK_SPEED;
			}
			else if(dx > -STRAFE_SPEED)
				dx -= STRAFE_ACCEL;
		}
		if(window.isKeyPressed(GLFW_KEY_S) && airborne) {
			if(dy > 0)
				fallSpeed = FASTFALL;
			dy += fallSpeed;
		}
		jump();
		updateAnim();
		if(dy < MAX_GRAV) {
			if(dy + GRAVITY > MAX_GRAV)
				dy = MAX_GRAV;
			else
				dy += GRAVITY;
		}
		x += dx;
		y += dy;
		anims[state].update();
		rHitbox.setPosition(x, y);
		if(!move) {
			if(!airborne && dx > 0) {
				dx -= GROUND_DECAY;
				if(dx < 0)
					dx = 0;
			}
			else if(!airborne && dx < 0) {
				dx += GROUND_DECAY;
				if(dx > 0)
					dx = 0;
			}
		}
		else if(dx > 0) {
			dx -= AERIAL_DECAY;
			if(dx < 0)
				dx = 0;
		}
		else if(dx < 0) {
			dx += AERIAL_DECAY;
			if(dx > 0)
				dx = 0;
		}
		if(window.isKeyPressed(GLFW_KEY_S) && airborne) {
			dy -= fallSpeed;
		}
	}
	
	public void render() {
		shader.setUniformMat4f("projection", camera.getProjection().translate(x, y, 0));
		anims[state].bind();
		shader.enable();
		if(dir == RIGHT)
			right.render();
		else if(dir == LEFT)
			left.render();
		shader.disable();
		anims[state].unbind();
	}
	
	private void jump() {
		if(airborne && jumps < 2)
			jumps = 2;
		if((window.isKeyPressed(GLFW_KEY_SPACE) || (jump > 0 && jump < 3)) && jumpFrames < MIN_JUMP) {
			if(jump == 0) {
				jump = 1;
				if(dx > 1)
					dx = 1;
				else if(dx < -1)
					dx = -1;
			}
			dy -= SHORT_HOP;
			airborne = true;
		}
		else if(window.isKeyPressed(GLFW_KEY_SPACE) && jumpFrames < MAX_JUMP)
			dy -= LONG_JUMP;
		if(jump % 2 == 1 && !window.isKeyPressed(GLFW_KEY_SPACE))
			++jump;
		else if(jump % 2 == 0 && jump > 1 && window.isKeyPressed(GLFW_KEY_SPACE) && jumps < NUM_JUMPS) {
			++jump;
			++jumps;
			dy = -AERIAL_JUMP;
			if(window.isKeyPressed(GLFW_KEY_A))
				dx = -1f;
			if(window.isKeyPressed(GLFW_KEY_D))
				dx = 1f;
		}
		if(jumps == 1)
			jumpFrames = MAX_JUMP;
		else if(jumpFrames < MAX_JUMP && jump > 0)
			++jumpFrames;
	}
	
	private void updateAnim() {
		int tAnim = 0;
		if(!airborne) {
			if(dx > 1 || dx < -1)
				tAnim = RUN;
			else if(dx > 0 || dx < 0)
				tAnim = WALK;
			else
				tAnim = IDLE;
			if(dx > 0)
				dir = RIGHT;
			else if(dx < 0)
				dir = LEFT;
		}
		else if(airborne && dy <= 0)
			tAnim = JUMP;
		else if(airborne && dy > 0)
			tAnim = FALL;
		if(tAnim != state) {
			anims[state].pause();
			anims[state].reset();
			anims[tAnim].play();
			state = tAnim;
		}
	}
	
	public void hitStage(boolean hit) {
		if(!hit)
			airborne = true;
		else {
			if(rHitbox.getX() != x) {
				if(Math.abs(dx) < BOUNCE_SPEED)
					dx = 0;
				else
					dx = -dx * BOUNCE_SCALE;
			}
			else if(rHitbox.getY() < y) {
				if(Math.abs(dy) < BOUNCE_SPEED) {
					fallSpeed = 0.2f;
					dy = 0;
					jump = 0;
					jumpFrames = 0;
					jumps = 0;
					airborne = false;
				}
				else
					dy = -dy * BOUNCE_SCALE;
			}
			else if(rHitbox.getY() > y) {
				if(Math.abs(dy) < BOUNCE_SPEED)
					dy = 0;
				else
					dy = -dy * BOUNCE_SCALE;
			}
			x = rHitbox.getX();
			y = rHitbox.getY();
		}
	}
	
	public RHitbox getRHitbox() {
		return rHitbox;
	}
}
